--[[
Import libraries, other files.
Save, load, clean and convert data.
Main window with shared elements.
Plugin's command.

DToX, 2013-2014
Licensed under the same terms as Lua itself.
--]]


--[[
Not indented comments - moving from v1.4 to v2.0 Tulkas database
--]]




-- import "Turbine";
-- import "Turbine.UI";
import "Turbine.UI.Lotro";
import "Turbine.Gameplay";

import "DToX.lib.lib";
local Lib = DToX.lib.Lib;
Lib.pName = "Tulkas-M"; -- For wl in lib file
fonts, colors, align = Lib.fonts, Lib.colors, Lib.align;
wl, getBorders, getPosition =
Lib.wl, Lib.getBorders, Lib.getPosition;
-- Plugin
Plugin = Lib.Plugin; -- Functions: convertData, language/lang, loadData, saveData
-- Tables
Table = Lib.Table; -- Functions: copy, dump, dupes, getKeys, length/len
dt = Table.dump;
-- Visual ones
addBorder = Lib.addBorder;
Hint = Lib.Hint; -- Functions: show, remove
-- Test ones
dv, mu = Lib.dv, Lib.memoryUsage;


-- Variables and constants
--
local pCommandName = "tm";
local pDataName = "TulkasM";
local pDataNameBackup = "TulkasM_Backup";
local pDataNameConverted = "TulkasM_Converted";
local pDataScope = Turbine.DataScope.Account;
data = { -- Contains all data
   ["Items"] = {};
   ["Sets"] = {};
   ["OutdatedNames"] = {};
   ["NotFound"] = {};
   ["Backup"] = true;
};
local newDatabase; -- Contains copy of v2.0 database
clickedCheckSets = false;
language = 0; -- For correct insertion of names
-- Main window, carrier and offsets size
local MAIN_CARRIER_WIDTH = 450;
local MAIN_CARRIER_HEIGHT = 50;
local OFFSET = 25;
local OFFSET_TOP = 40;
local MAIN_WINDOW_WIDTH = MAIN_CARRIER_WIDTH + OFFSET * 2;
local MAIN_WINDOW_HEIGHT = MAIN_CARRIER_HEIGHT + OFFSET + OFFSET_TOP;


-- Forward declarations
--
local btn_items;
local btn_sets;
local mainCarrierItems;
local mainCarrierSets;
local windowVeil;
-- Returned from MItems.lua
local btn_updateNames;
local btn_removeItems;
local txt_result;
-- Returned from MSets.lua
local btn_checkSets;





-- Functions
--
-- Tries to load Tulkas DB. Returns true or false. Shows messages
local function loadDB( name )
   wl( ">> Outsourcing to Russia... Loading DB (" .. name .. ")..." );

   -- From Tulkas (Eng DB; full). Array's name: _ITEMSDB
   local ok = pcall( import, "DToX.bases." .. name );
   -- local ok = pcall( import, "DToX.TulkasToolkit.Items" );

   _ITEMSDB = _ITEMSDB or DToX.bases._ITEMSDB;
   if ( ok ) then
      if ( not _ITEMSDB ) then
         wl( ">> Tulkas' '_ITEMSDB' variable is not found.", "Error" );
         return false;
      else
         local next = next;
         if ( not next( _ITEMSDB ) ) then
            wl( ">> Tulkas' '_ITEMSDB' variable is empty.", "Error" );
            return false;
         else
            wl( ">> DB is ready for torture... Items: " .. Table.len( _ITEMSDB )  .. "." );
            return true;
         end
      end
   else
      wl( ">> No Tulkas' '" .. name .. ".lua'. Please, copy it to 'bases' folder.", "Error" );
      return false;
   end
end

-- Shows corresponding tab with visual elements, hides other
local function changeTab( name )
   btn_items:SetEnabled( false );
   btn_sets:SetEnabled( false );
   mainCarrierItems:SetVisible( false );
   mainCarrierSets:SetVisible( false );

   if ( name == "Items" ) then
      btn_sets:SetEnabled( true );
      mainCarrierItems:SetVisible( true );
   elseif ( name == "Sets" ) then
      btn_items:SetEnabled( true );
      mainCarrierSets:SetVisible( true );
   end
end

-- Loads additional data from Tulkas DB.
-- Converts data to Tulkas format and saves it to file
local function convertData()
   -- Check, if Tulkas DB variable is available. If not, try to load
   if ( not _ITEMSDB ) then
      local ok = pcall( import, "DToX.bases.Items" );
      if ( not ok ) then
         wl( ">> No Tulkas' 'Items.lua'. Can't convert.", "Error" );
         return;
      end
      _ITEMSDB = _ITEMSDB or DToX.bases._ITEMSDB;
   end

   -- Check, what database is loaded
   if ( newDatabase ) then
      _ITEMSDB = newDatabase;
   else
      for ID, v in pairs( _ITEMSDB ) do
         if ( v.Name ) then
            wl( "No v2.0 database. Please, save data, unload plugin, load plugin and try to convert data again.", "Error" );
         end
         return;
      end
   end

   -- Adds quotes around string values
   local type = type;
   local items = data.Items;
   local total = 0;
   local database = {}; -- All items

   local function addQuotes( s )
      if ( type( s ) == "string" ) then
         return "\"" .. s .. "\"";
      else
         return s;
      end
   end

   -- Converts table to string
   local function tableToString( t, mode )
      local line = "";

      for k, v in pairs( t ) do
         local k = addQuotes( k );
         if( k == 15 ) then
            mode = "IDs";
         end

         line = line .. "[" .. k .. "]=";

         -- IDs
         if ( mode == "IDs" and k == 2 and type( v ) == "table" ) then
            line = line .. "{" .. table.concat( t[k], "," ) .. "};"
            mode = nil;
         else
            if ( type( v ) == "table" ) then
               line = line .. tableToString( v, mode );
               -- Stats
               if ( k == 6 ) then
                  if ( line:find( "," ) ) then
                     line = line:gsub( ",", "." );
                  end
               end
            else
               local v = addQuotes( v );
               line = line .. v .. ";";
            end
         end

      end

      line = "{" .. line .. "};"
      return line;
   end


   -- Get additional items' data from Tulkas DB
   for ID in pairs( items ) do
      local numID = tonumber( ID );
      for k, v in pairs( _ITEMSDB[numID] ) do
-- if ( k ~= "Name" and k ~= "Set" ) then
         if ( k ~= 1 and k ~= 15 ) then
            items[ID][k] = v;
         end
      end
   end

   -- Convert items' data to strings
   local sortedKeys = Table.getKeys( items, true );
   for _, ID in pairs( sortedKeys ) do
      v = items[ID];
      local line = ""; -- For one item

      local name = v[1][language];
      -- Correct name. Replace "’" with "'"
      if ( name:find( "s’ " ) ) then
         name = name:gsub( "s’ ", "s' " );
      end
      if ( name:find( "’s" ) ) then
         name = name:gsub( "’s", "'s" );
      end
      if ( name:find( "�s" ) ) then
         name = name:gsub( "�s", "'s" );
      end
      v[1][language] = name;

      line = "[" .. ID .. "]=" .. tableToString( v );

      -- Prepare for saving
      if ( line:find( "'" ) ) then
         line = line:gsub( "'", "__" );
      end
      line = line:gsub( '"', "#" );

      local j = #database + 1;
      database[j] = line;
      total = total + 1;
   end

   saveData( pDataNameConverted, database );
   wl( ">> Converted items: " .. total .. "."
   .. "\nSaved to file ...\\<Documents>\\The Lord of the Rings Online\\PluginData\\<YourLogin>\\AllServers\\" .. pDataNameConverted .. ".plugindata."
   .. "\nUse supplied *.html file to clean it and get Tulkas DB." );
end

-- Removes raw sets, corrects indexes, returns result
local function cleanData()
   local cData = Table.copy( data );
   -- Remove raw sets, correct indexes
   local tmp = {};
   for i, v in pairs( cData.Sets ) do
      if ( not v.isRaw ) then
         local j = #tmp + 1;
         tmp[j] = v;
      end
   end
   cData.Sets = tmp;

   -- Remove all fields except Name and Set
   local tmp = {};
   for ID, info in pairs( cData.Items ) do
      tmp[ID] = {
         [1] = info[1];
         [15] = info[15];
      };
   end
   cData.Items = tmp;

   cData = Plugin.convertData( cData, "To" );

   return cData;
end

local function loadBackup( savedData )
   if ( savedData ) then
      wl( ">> Loaded backup." );
      data = Plugin.convertData( savedData, "From" );
   else
      wl( ">> Backup is empty ¯\\(º_o)/¯", "Error" );
   end

   -- Update lists with sets and hide veil
   if ( clickedCheckSets ) then
      btn_checkSets.MouseClick();
   end
   windowVeil:SetVisible( false );
   windowVeil:SetParent( nil );
end

-- Adds shortcut to quickslot, returns it
function addShortcut( quickslot, decID )
   local hex = string.format( "%X", decID );
   local sc = Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Item, "0x0,0x" .. hex );
   quickslot:SetShortcut( sc );
   return sc;
end

-- Creates and returns object with specified and predifined properties
function newObject( parent, mode, width, text, fontSize )
   local fontMap = {
      [12] = fonts.Verdana12;
      [14] = fonts.Verdana14;
   };
   local fontSize = fontSize or 14; -- Default to 14, if nil
   local font = fontMap[fontSize] or fontSize; -- If user specified font

   if ( mode == "Button" ) then
      local button = Turbine.UI.Lotro.Button();
      button:SetParent( parent );
      button:SetFont( font );
      button:SetText( text );
      button:SetSize( width, 20 );
      return button;

   elseif ( mode == "Label" ) then
      local label = Turbine.UI.Label();
      label:SetParent( parent );
      label:SetFont( font );
      label:SetText( text );
      label:SetSize( width, 20 );
      return label;

   elseif ( mode == "Text" ) then
      local text = Turbine.UI.Lotro.TextBox();
      text:SetParent( parent );
      text:SetFont( font );
      text:SetSize( width, 20 );
      return text;

   end
end

function saveData( name, data )
   Turbine.PluginData.Save( pDataScope, name, data );
end

function loadData( name, onLoad )
   return Turbine.PluginData.Load( pDataScope, name, onLoad );
end

-- Clears data and saves backup
function Turbine.Plugin.Unload()
   if ( data.Backup ) then
      local data = cleanData();
      saveData( pDataNameBackup, data );
      wl( ">> Saved backup." );
   end
end




-- Visuals
--

-- Main window
win_main = Turbine.UI.Lotro.Window();
win_main:SetText( "Madness of... Tulkas" );
win_main:SetVisible( true );
win_main:SetSize( MAIN_WINDOW_WIDTH, MAIN_WINDOW_HEIGHT );
local left = 200;
local top = 100;
win_main:SetPosition( left, top );
function win_main:Closed()
   wl( "To show window use command: /tm or /tm show." );
end
-- Resize veil
function win_main:SizeChanged()
   local w, h = win_main:GetSize();
   windowVeil:SetSize( w, h );
end

-- Main carrier for elements (offseted from window's borders)
local mainCarrier = Turbine.UI.Control();
mainCarrier:SetParent( win_main );
mainCarrier:SetSize( MAIN_CARRIER_WIDTH, MAIN_CARRIER_HEIGHT );
mainCarrier:SetPosition( OFFSET, OFFSET_TOP );

-- Main carriers for elements from other files
local b = getBorders( mainCarrier );
local left = b.Left;
local top = b.Bottom + 20;
mainCarrierItems = Turbine.UI.Control();
mainCarrierItems:SetParent( win_main );
mainCarrierItems:SetPosition( left, top );
mainCarrierSets = Turbine.UI.Control();
mainCarrierSets:SetParent( win_main );
mainCarrierSets:SetPosition( left, top );

-- Button: Work with items
btn_items = newObject( mainCarrier, "Button", 130, "Work with items" );
btn_items:SetPosition( 0, 0 );
-- Show tab, load file, return elements for manipulation
function btn_items:MouseClick()
   changeTab( "Items" );
   import "DToX.TulkasToolkit.MItems";
   txt_result, btn_updateNames, btn_removeItems = showItems( mainCarrierItems );
end

-- Button: Work with sets
btn_sets = newObject( mainCarrier, "Button", 130, "Work with sets" );
local b = getBorders( btn_items );
btn_sets:SetPosition( b.Right + 30, b.Top );
-- Show tab, load file, return elements for manipulation
function btn_sets:MouseClick()
   changeTab( "Sets" );
   import "DToX.TulkasToolkit.MSets";
   btn_checkSets = showSets( mainCarrierSets );
end

-- Button: Convert data
local btn_convertData = newObject( mainCarrier, "Button", 130, "Convert data" );
local b = getBorders( btn_sets );
btn_convertData:SetPosition( b.Right + 30, b.Top );
function btn_convertData:MouseHover()
   Hint.show( self, "Convert current data to Tulkas format. Result'll be saved in file. Then use supplied *.html file to clean it and get Tulkas DB.\n\nWarning! Takes some tome. Game may not respond during conversion." );
end
-- Check sets, load additional data from Tulkas DB, convert
function btn_convertData:MouseClick()
   -- Check sets
   if ( not btn_checkSets ) then
      btn_sets.MouseClick();
      btn_items.MouseClick();
      btn_checkSets.MouseClick();
   end

   convertData();
end

-- Button: Load DB
local btn_loadDB = newObject( mainCarrier, "Button", 130, "Load database" );
local b = getBorders( btn_items );
btn_loadDB:SetPosition( b.Left, b.Bottom + 10 );
function btn_loadDB:MouseHover()
   Hint.show( self, "Load data from Tulkas Items.lua file.\n\nWarning! It'll rewrite current items." );
end
-- Load DB or show message, prepare elements on Items tab, re-check sets
function btn_loadDB:MouseClick()
   if ( loadDB( "Items" ) ) then
      data.Items = Table.copy( _ITEMSDB );
      newDatabase = Table.copy( _ITEMSDB );
      _ITEMSDB = nil;
      btn_loadDB:SetEnabled( false );
   end

   if ( btn_updateNames ) then
      btn_updateNames:SetEnabled( false );
      btn_removeItems:SetEnabled( false );
      txt_result:SetText( "" );
   end

   if ( clickedCheckSets ) then
      btn_checkSets.MouseClick();
   end
end

-- Button: Import names from v1.4 DB
local btn_importOld = newObject( mainCarrier, "Button", 130, "Import from v1.4" );
local b = getBorders( btn_sets );
btn_importOld:SetPosition( b.Left, b.Bottom + 10 );
function btn_importOld:MouseHover()
   Hint.show( self, "Load names (items and sets) from Tulkas v1.4 database file.\n\nFile must be named 'ItemsOld.lua'.\n\nLoads new database first." );
end
function btn_importOld:MouseClick()
   local notLoadedNewDB = btn_loadDB:IsEnabled();
   if ( notLoadedNewDB ) then
      btn_loadDB.MouseClick();
   end

   if ( loadDB( "ItemsOld" ) ) then
      btn_importOld:SetEnabled( false );
      local oldBase = _ITEMSDB;
      local newBase = data.Items;

      -- Insert old names into new database
      for numID, info in pairs( oldBase ) do
         local itemName = info.Name;
         local setName;
         if ( info.Set ) then
            setName = info.Set.Name;
         end
         if ( newBase[numID] ) then
            newBase[numID][1][language] = itemName or newBase[numID][1][language];
            if ( newBase[numID][15] and setName ) then
               newBase[numID][15][1][language] = setName;
            end
         end
      end
      wl( ">> Imported translated names from old database." );
   end
end


-- Button: Save data
local btn_saveData = newObject( mainCarrier, "Button", 130, "Save data" );
local b = getBorders( btn_convertData );
btn_saveData:SetPosition( b.Left, b.Bottom + 10 );
-- Clean sets, save data, show path to file
function btn_saveData:MouseClick()
   local data = cleanData();
   saveData( pDataName, data ); -- !! Improve time?
   wl( ">> Saved to file ...\\<Documents>\\The Lord of the Rings Online\\PluginData\\<YourLogin>\\AllServers\\" .. pDataName .. ".plugindata." );
end

-- Veil: lies over window
windowVeil = newObject( win_main, "Label", 130, "Loading...", fonts.Verdana23 );
windowVeil:SetBackColor( colors.Grey05 );
windowVeil:SetBackColorBlendMode( 3 );
windowVeil:SetTextAlignment( align[5] );
windowVeil:SetVisible( false );





-- Add command
--
pCommand = Turbine.ShellCommand();
Turbine.Shell.AddCommand( pCommandName, pCommand );
function pCommand:Execute( cmd, args )
   if ( args == "" or args == "show" ) then
      win_main:SetVisible( true );
      win_main:Activate();

   elseif ( args == "?" ) then
      local c = "/" .. cmd;
      wl (
      ">> Available commands:"
      .. "\n" .. c .. " ? - shows this message."
      .. "\n" .. c .. " or " .. c .. " show - shows window."
      .. "\n" .. c .. " backup - enables or disables backup."
      .. "\n" .. c .. " load backup - loads backup, which is saved when you unload plugin."
      .. "\n" .. c .. " clear data - removes all items and sets. If something went wrong and you don't want to wait for backup to be saved. Or just don't want to wait and all data was saved manually."
      .. "\n" .. c .. " clear items - removes all items."
      .. "\n" .. c .. " clear sets - removes all sets."
      .. "\n" .. c .. " mcs - marks all sets, that have bonuses or translated names as edited."
      .. "\n" .. c .. " mcs bonus - marks all sets, that have bonuses as edited."
      .. "\n" .. c .. " mcs name - marks all sets, that have translated names as edited."
      .. ""
      );

   elseif ( args == "backup" ) then
      data.Backup = not data.Backup;
      wl( ">> Backup is enabled: ", data.Backup, "." );

   elseif ( args == "load backup" ) then
      loadData( pDataNameBackup, loadBackup );
      wl( ">> Loading backup... please, wait ~15 seconds." );
      windowVeil:SetParent( win_main );
      windowVeil:SetVisible( true );

   elseif ( args == "clear data" ) then
      data.Items = {};
      data.Sets = {};
      wl( ">> All items and sets are removed. To reverse it, unload and re-load plugin. Don't press 'Save data'.", "Warning" );
      if ( clickedCheckSets ) then
         btn_checkSets.MouseClick();
      end

   elseif ( args == "clear sets" ) then
      data.Sets = {};
      wl( ">> All saved sets are removed. To reverse it, unload plugin. Don't press 'Save data'.", "Warning" );
      if ( clickedCheckSets ) then
         btn_checkSets.MouseClick();
      end

   elseif ( args == "clear items" ) then
      data.Items = {};
      wl( ">> All items are removed." );

   -- Moves sets, that have filled Set.Bonus key to the edited list
   -- For database from M-13 Project
   elseif ( args == "mcs" or args == "mcs bonus" or args == "mcs name" ) then
      local argsMap = {
         ["mcs"] = " >> Sets with bonuses or translated names moved to edited list.";
         ["mcs bonus"] = ">> Sets with bonuses moved to edited list.";
         ["mcs name"] = ">> Sets with translated names moved to edited list.";
      };
   
      -- Load tab and create set's list
      if ( not clickedCheckSets ) then
         btn_sets.MouseClick();
         btn_checkSets.MouseClick();
      end

      local sets = data.Sets;
      local items = data.Items;
      for setIndex, setData in pairs( sets ) do
         local numID = setData.ID;
         local item = items[numID];
         local bMove = false;

-- if ( item.Set and item.Set.Bonus ) then
         if ( item[15] ) then

            if ( ( args == "mcs" or args == "mcs bonus" ) and item[15][3] ) then
-- local setBonusesNum = Table.len( item.Set.Bonus );
               local setBonusesNum = Table.len( item[15][3] );
               if ( setBonusesNum > 0 ) then
                  bMove = true;
               end
            end

            if ( ( args == "mcs" or args == "mcs name" ) and item[15][1][language] ~= "" ) then
               bMove = true;
            end


            if ( bMove ) then
-- sets[setIndex].Set = item.Set;
               sets[setIndex][15] = item[15];
               sets[setIndex].isRaw = false;
            end

         end -- If has set field

      end

      -- Update sets' lists
      if ( clickedCheckSets ) then
         btn_checkSets.MouseClick();
      end

      wl( argsMap[args] );

   else
      wl( "Error. Please, check command.", "Error" );
   end
end





-- Logic
--

-- Open items tab
btn_items.MouseClick();
-- btn_sets.MouseClick();

-- Determine language
local languageCodeName = Plugin.lang();
local languageMap = {
   ["ENG"] = 1;
   ["FRE"] = 2;
   ["GER"] = 3;
   ["RUS"] = 4;
};
language = languageMap[languageCodeName];

-- Load saved data
local savedData = loadData( pDataName );
if ( savedData ) then
   wl( ">> Loaded saved data." );
   data = Plugin.convertData( savedData, "From" );
else
   wl( ">> No saved data." );
   btn_loadDB.MouseClick();
end

if ( data.Backup == nil ) then
   data.Backup = true;
   data.Items = {};
   data.Sets = {};
end

wl( ">> Save data for language: " .. languageCodeName .. ".", "Warning" );
wl( ">> Backup is enabled: ", data.Backup, "." );

wl( "\nWelcome. Don't forget to press 'Save data', when you're done."
.. "\nTo remove all saved sets use command: /tm clear sets."
.. "\nTo load backup use command: /tm load backup. Backup is saved, when you unload the plugin."
.. "\nTo show available commands use command: /tm ?."
);